{% extends "base.html" %}

{% block content %}
<div class="container position-absolute top-50 start-50 translate-middle col-md-8 col-12"
  style="max-height: 80vh; overflow-y: auto;">
  <h2>Deferred Tasks</h2>
  <table class="table table-striped">
    <thead>
      <tr>
        <th scope="col">Task ID</th>
        <th scope="col">Scheduled Time</th>
      </tr>
    </thead>
    <tbody>
      {% for task_id, dt in deferred.items() %}
      <tr>
        <td>{{ task_id }}</td>
        <td>{{ dt }}</td>
      </tr>
      {% endfor %}
    </tbody>
  </table>
  <h2>Enqueued Tasks</h2>
  <table class="table table-striped">
    <thead>
      <tr>
        <th scope="col">Task ID</th>
        <th scope="col">Enqueue Time</th>
      </tr>
    </thead>
    <tbody>
      {% for task_id, dt in enqueued.items() %}
      <tr>
        <td>{{ task_id }}</td>
        <td>{{ dt }}</td>
      </tr>
      {% endfor %}
    </tbody>
  </table>
</div>
{% endblock %}
