{% extends "base.html" %}

{% block content %}
<div class="container position-absolute top-50 start-50 translate-middle col-md-8 col-12">
  <table class="table table-striped">
    <thead>
      <tr>
        <th scope="col">Queue</th>
        <th scope="col">Active</th>
        <th scope="col">Enqueued</th>
        <th scope="col">Scheduled</th>
        <th scope="col">Workers</th>
      </tr>
    </thead>
    <tbody>
      {% for queue in queues %}
      <tr>
        <td><a href="/queues/{{ queue['name'] }}">{{ queue['name'] }}</a></td>
        <td>{{ queue['active'] }}</td>
        <td>{{ queue['enqueued'] }}</td>
        <td>{{ queue['scheduled'] }}</td>
        <td>{{ queue['workers'] }}</td>
      </tr>
      {% endfor %}
    </tbody>
  </table>
</div>
{% endblock %}
